/**************
 * Copyright (c) brent@meshier.com 2006
 *************/

#include "local_scan.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mysql/mysql.h>

/************************/
#define MYSQLHOST "localhost"
#define MYSQLUSER "greylist"
#define MYSQLPASS "topSecret"
#define MYSQLDB   "qmail"
#define BLOCK_EXPIRE 5 /* minutes until email is accepted */
#define RECORD_EXPIRE 1500 /* minutes until record expires */
#define RECORD_EXPIRE_GOOD 36 /* days until record expires after accepting email */

/*************************/

#define DBGSCAN 1
#define SQLCMDSIZE 1024

FILE *dbglog = NULL;

int mysql_query_wrapper(MYSQL *mysql, char *sqltext) 
{
  int result;

  result = mysql_query(mysql,sqltext);
/*  if (dbglog)                                              */
/*    fprintf(dbglog,"SQL: ret=%d  |%s|\n",result,sqltext);  */
  return result;
}

void AddOrClause(char *subquery,char *in_ipaddr)
{
 int loop;
 char *ptr;
 char ipaddr[256];

 strncpy(ipaddr, in_ipaddr, sizeof(ipaddr));
 for (loop = 0; loop < 4; loop++) {
   if (loop)
     strcat(subquery, " OR ");
   strcat(subquery , "relay_ip = '" );
   strcat(subquery ,ipaddr);
   strcat(subquery , "'");
   ptr = strrchr(ipaddr,'.'); // strip off the last octet
   if(ptr)
     *ptr = '\0';
 }
}

/* returns TRUE FALSE if record exists in database */
int checkWhiteListIP( MYSQL *mysql,  int *action) 
{
  MYSQL_RES *myres; 
  MYSQL_ROW myrow;

  char sql[SQLCMDSIZE], sid[32];
  int  white, black,exists = FALSE;

  /* check for whitelisted sender ip address */
     sprintf(sql, "SELECT id, block_expires > NOW(), block_expires < NOW() FROM relaytofrom WHERE record_expires > NOW()  AND mail_from IS NULL AND rcpt_to IS NULL AND (");
    AddOrClause(sql, sender_host_address);
    strcat(sql, ") ORDER BY length(relay_ip)");
    //#ifdef DBGSCAN
    //   fprintf(dbglog,"check whitelist: %s\n",sql);
    //#endif
    if(!mysql_query_wrapper(mysql,sql)) {
      myres = mysql_store_result(mysql);
      while ((myrow = mysql_fetch_row(myres)) != NULL) {
	exists = TRUE;
	strncpy(sid,myrow[0],sizeof(sid));
	black = atoi(myrow[1]);
	white = atoi(myrow[2]);
      }
      mysql_free_result(myres); // free memory used by result
    }
    if (exists && white) {
#ifdef DBGSCAN
      fprintf(dbglog,"%s -> xxx (%s) Whitelist IP Accept id = %s  \n",sender_address, sender_host_address,sid);
#endif
      *action = LOCAL_SCAN_ACCEPT;
    }
    else if (exists && black) {
#ifdef DBGSCAN
      fprintf(dbglog,"%s -> xxx (%s) Blacklist IP Permanent Reject id = %s  \n",sender_address, sender_host_address,sid);
#endif
      *action = LOCAL_SCAN_REJECT;
    }

  return exists;
}


/* returns TRUE FALSE if record exists in database */
int checkWhiteListRcpt( MYSQL *mysql,  int *action, int i) 
{
  MYSQL_RES *myres; 
  MYSQL_ROW myrow;

  char sql[SQLCMDSIZE], sid[32];
  int  white, black,exists = FALSE;

  if(recipients_list[i].address != NULL) {

    /* check for whitelisted receiver domain address */
    sprintf(sql, "SELECT id, block_expires > NOW(), block_expires < NOW() FROM relaytofrom WHERE record_expires > NOW() AND mail_from IS NULL AND relay_ip IS NULL AND rcpt_to = '%s'", recipients_list[i].address);
    if(!mysql_query_wrapper(mysql,sql)) {
      myres = mysql_store_result(mysql);
      while ((myrow = mysql_fetch_row(myres)) != NULL) {
	exists = TRUE;
	strncpy(sid,myrow[0],sizeof(sid));
	black = atoi(myrow[1]);
	white = atoi(myrow[2]);
      }
      mysql_free_result(myres); // free memory used by result
    }

    if (exists && white) {
#ifdef DBGSCAN
      fprintf(dbglog,"%s -> %s (%s) Whitelist Domain Accept id = %s  \n",sender_address, recipients_list[i].address, sender_host_address,sid);
#endif
      *action = LOCAL_SCAN_ACCEPT;
    }
    else if (exists && black) {
#ifdef DBGSCAN
      fprintf(dbglog,"%s -> xxx (%s) Blacklist Domain Permanent Reject id = %s  \n",sender_address, sender_host_address,sid);
#endif
      *action = LOCAL_SCAN_REJECT;
    }

  }
  return exists;
}

int checkWhiteListFrom( MYSQL *mysql,  int *action) 
{
  MYSQL_RES *myres; 
  MYSQL_ROW myrow;

  char sql[SQLCMDSIZE], sid[32];
  int  white, black,exists = FALSE;

  if(sender_address != NULL) {

    /* check for whitelisted sender address */
    sprintf(sql, "SELECT id, block_expires > NOW(), block_expires < NOW() FROM relaytofrom WHERE record_expires > NOW() AND rcpt_to IS NULL AND relay_ip IS NULL AND mail_from = '%s'", sender_address);
    if(!mysql_query_wrapper(mysql,sql)) {
      myres = mysql_store_result(mysql);
      while ((myrow = mysql_fetch_row(myres)) != NULL) {
	exists = TRUE;
	strncpy(sid,myrow[0],sizeof(sid));
	black = atoi(myrow[1]);
	white = atoi(myrow[2]);
      }
      mysql_free_result(myres); // free memory used by result
    }

    if (exists && white) {
#ifdef DBGSCAN
      fprintf(dbglog,"%s (%s) Whitelist From Address Accept id = %s  \n",sender_address, sender_host_address,sid);
#endif
      *action = LOCAL_SCAN_ACCEPT;
    }
    else if (exists && black) {
#ifdef DBGSCAN
      fprintf(dbglog,"%s -> xxx (%s) Blacklist From Address Permanent Reject id = %s  \n",sender_address, sender_host_address,sid);
#endif
      *action = LOCAL_SCAN_REJECT;
    }
  }
  return exists;
}

void  buildLookupSql(char *sql, int recipIndex) 
{
  char hostip[32];
  char *ptr;

  if (FALSE) {
    /* this first sprintf does an exact match on ipaddr */
    sprintf(sql,"SELECT id, NOW() > block_expires FROM relaytofrom WHERE record_expires > NOW() AND mail_from = '%s'  AND rcpt_to   = '%s' AND relay_ip  = '%s' order by block_expires desc",sender_address,  recipients_list[recipIndex].address,sender_host_address);
  }
  else {
    /* this second version matches anything in the same /24 subnet */
    sprintf(sql,"SELECT id, NOW() > block_expires FROM relaytofrom WHERE record_expires > NOW() AND mail_from = '%s'  AND rcpt_to   = '%s' AND relay_ip  like '",sender_address,  recipients_list[recipIndex].address);
    strcpy(hostip,sender_host_address);
    /* now remove the last octet and add a '%' */
    ptr = strrchr(hostip,'.'); // strip off the last octet    
    if(ptr)
      *ptr = '\0';
    strcat(sql,hostip);
    strcat(sql, "%' order by block_expires desc");
  }
}


int checkGreylist( MYSQL *mysql,  int *action, int i) 
{
 MYSQL_RES *myres; 
 MYSQL_ROW myrow;

 char sql[SQLCMDSIZE], sid[32];
 int notexpired, exists = FALSE;

 // lookup to see if record exists
 buildLookupSql(sql,i);
 if(!mysql_query_wrapper(mysql,sql)) {
   myres = mysql_store_result(mysql);
   while ((myrow = mysql_fetch_row(myres)) != NULL) {
     exists = TRUE;
     strncpy(sid,myrow[0],sizeof(sid));
     notexpired = atoi(myrow[1]);
     //fprintf(dbglog,"id = %s  expire = %d\n",sid,notexpired);
   }
   mysql_free_result(myres); // free memory used by result
   // if it exists and is not record expired and the block_expired passes
   if (exists && notexpired) {
     // update expire time to 36 days, and pass count
     sprintf(sql,"update relaytofrom set record_expires = NOW() + INTERVAL %d DAY, passed_count = passed_count + 1 where id ='%s'",RECORD_EXPIRE_GOOD,sid);
     mysql_query_wrapper(mysql, sql);
#ifdef DBGSCAN
     fprintf(dbglog,"%s -> %s (%s) Exists Accept id = %s  expire = %d\n",sender_address,recipients_list[i].address, sender_host_address,sid,notexpired);
#endif
     *action = LOCAL_SCAN_ACCEPT;
   }
   else if ( exists && ! notexpired) {
     // update fail count
     sprintf(sql,"update relaytofrom set blocked_count = blocked_count + 1 where id = %s",sid);
     mysql_query_wrapper(mysql, sql);
#ifdef DBGSCAN
     fprintf(dbglog,"%s -> %s (%s) Exists Block id = %s  expire = %d\n",sender_address,recipients_list[i].address, sender_host_address,sid,notexpired);
#endif
     *action = LOCAL_SCAN_TEMPREJECT;
   }
   else /* doesn't exist */ {
     // it doesn't exist make and entry
     sprintf(sql,"insert into relaytofrom values (0,'%s','%s','%s',NOW() + INTERVAL %d MINUTE,NOW() + INTERVAL %d MINUTE,1,0,0,'AUTO',NOW(),NOW())",sender_host_address,sender_address,  recipients_list[i].address,BLOCK_EXPIRE,RECORD_EXPIRE);
     mysql_query_wrapper(mysql,sql);
#ifdef DBGSCAN
     fprintf(dbglog,"%s -> %s (%s) Doesn't Exists Block\n",sender_address,recipients_list[i].address, sender_host_address);
#endif
     *action = LOCAL_SCAN_TEMPREJECT;
   }
 }
}


int local_scan(int fd, uschar **return_text)
{
 MYSQL *mysql = NULL;
 int i,  ret = LOCAL_SCAN_ACCEPT;



#ifdef DBGSCAN
 dbglog = fopen("/tmp/greylist_dbg.txt","a");
 fprintf(dbglog,"protocol = %s %s  \n",received_protocol,sender_address);
#endif


 fd = fd;                      /* Keep picky compilers happy */
 return_text = return_text;                     /* Keep picky compilers happy */
 mysql = mysql_init(NULL);
 if (mysql && strcmp(received_protocol,"local") ) {
   if (mysql_real_connect(mysql,MYSQLHOST,MYSQLUSER,MYSQLPASS,MYSQLDB,0,NULL,0)) {
     if ( !checkWhiteListIP( mysql, &ret )) {      /* check for whitelisted sender ip address */
       if ( !checkWhiteListFrom( mysql, &ret )) {    /* check for whitelisted from address */
       for(i= 0 ; i <  recipients_count; i++ ) {
	 if(strlen(sender_host_address) + strlen(sender_address) +
	    strlen(recipients_list[i].address) < SQLCMDSIZE - 200) { /* check to avoid buffer overflows */
	   if(!checkWhiteListRcpt( mysql, &ret,i ))
	     checkGreylist(mysql, &ret,i);
	 }
        }
       }
     }
   }
 }
 if(mysql) mysql_close(mysql);
 if(dbglog) {
   fprintf(dbglog, "--------\n");
   fclose(dbglog);
 }
 return ret;
}

/* End of local_scan.c */
