#include "sig.h"
#include "readwrite.h"
#include "stralloc.h"
#include "substdio.h"
#include "alloc.h"
#include "auto_qmail.h"
#include "control.h"
#include "received.h"
#include "constmap.h"
#include "error.h"
#include "ipme.h"
#include "ip.h"
#include "qmail.h"
#include "str.h"
#include "fmt.h"
#include "scan.h"
#include "byte.h"
#include "case.h"
#include "env.h"
#include "now.h"
#include "exit.h"
#include "fork.h"
#include <stdlib.h>
#include <string.h>
#include "rcpthosts.h"
#include "timeoutread.h"
#include "timeoutwrite.h"
#include "commands.h"
#include "wait.h"
#include "fd.h"
#include "spf.h"
#include "qmail-spp.h"

int spp_val;

#define USE_SMTPAUTH
//#define SMTPAUTH_NETSCAPE_WORKAROUND
#define USE_OLD_GREETING
#define USE_NEW_GREETING

#define AUTHCRAM
#define MAXHOPS 100
unsigned int databytes = 0;
int timeout = 1200;
unsigned int spfbehavior = 0;

const char *protocol = "SMTP";

#ifdef TLS
# include "tls.h"
# include "ssl_timeoutio.h"

void tls_init();
int tls_verify();
void tls_nogateway();
void smtp_tls(char* arg);
int ssl_rfd = -1, ssl_wfd = -1; /* SSL_get_Xfd() are broken */
#endif

int safewrite(fd,buf,len) int fd; char *buf; int len;
{
  int r;
#ifdef TLS
  if (ssl && fd == ssl_wfd)
    r = ssl_timeoutwrite(timeout, ssl_rfd, ssl_wfd, ssl, buf, len);
  else
#endif
  r = timeoutwrite(timeout,fd,buf,len);
  if (r <= 0) _exit(1);
  return r;
}

char ssoutbuf[512];
substdio ssout = SUBSTDIO_FDBUF(safewrite,1,ssoutbuf,sizeof ssoutbuf);

void flush() { substdio_flush(&ssout); }
void out(s) const char *s; { substdio_puts(&ssout,s); }

void die_read() { _exit(1); }
void die_alarm() { out("451 timeout (#4.4.2)\r\n"); flush(); _exit(1); }
void die_nomem() { out("421 out of memory (#4.3.0)\r\n"); flush(); _exit(1); }
void die_control() { out("421 unable to read controls (#4.3.0)\r\n"); flush(); _exit(1); }
void die_ipme() { out("421 unable to figure out my IP addresses (#4.3.0)\r\n"); flush(); _exit(1); }
void straynewline() { out("451 See http://pobox.com/~djb/docs/smtplf.html.\r\n"); flush(); _exit(1); }

void err_tempfail() { out("421 temporary envelope failure (#4.3.0)\r\n"); }
void err_permfail() { out("553 sorry, permanent envelope failure (#5.7.1)\r\n"); }
void err_bmf() { out("553 sorry, your envelope sender is in my badmailfrom list (#5.7.1)\r\n"); }
#ifndef TLS
void err_nogateway() { out("553 sorry, that domain isn't in my list of allowed rcpthosts (#5.7.1)\r\n"); }
#else
void err_nogateway()
{
  out("553 sorry, that domain isn't in my list of allowed rcpthosts");
  tls_nogateway();
  out(" (#5.7.1)\r\n");
}
#endif
void err_unimpl(char* arg) { out("502 unimplemented (#5.5.1)\r\n"); }
void err_syntax() { out("555 syntax error (#5.5.4)\r\n"); }
void err_relay() { out("553 we don't relay (#5.7.1)\r\n"); }
void err_wantmail() { out("503 MAIL first (#5.5.1)\r\n"); }
void err_wantrcpt() { out("503 RCPT first (#5.5.1)\r\n"); }
void err_noop(char* arg) { out("250 ok\r\n"); }
void err_vrfy(char* arg) { out("252 send some mail, i'll try my best\r\n"); }
void err_qqt() { out("451 qqt failure (#4.3.0)\r\n"); }


stralloc greeting = {0};
stralloc spflocal = {0};
stralloc spfguess = {0};
stralloc spfexp = {0};

void smtp_greet(code) char *code;
{
  substdio_puts(&ssout,code);
  substdio_put(&ssout,greeting.s,greeting.len);
}
void smtp_help(char* arg)
{
  out("214 qmail home page: http://pobox.com/~djb/qmail.html\r\n");
}
void smtp_quit(char* arg)
{
  smtp_greet("221 "); out("\r\n"); flush(); _exit(0);
}

char *remoteip;
char *remotehost;
char *remoteinfo;
char *local;
char *relayclient;
#ifdef USE_SMTPAUTH
char *plesk_smtp_auth;
#endif

stralloc helohost = {0};
char *fakehelo; /* pointer into helohost, or 0 */

void dohelo(arg) char *arg; {
  if (!stralloc_copys(&helohost,arg)) die_nomem(); 
  if (!stralloc_0(&helohost)) die_nomem(); 
  fakehelo = case_diffs(remotehost,helohost.s) ? helohost.s : 0;
}

int liphostok = 0;
stralloc liphost = {0};
int bmfok = 0;
stralloc bmf = {0};
struct constmap mapbmf;

#ifdef USE_SMTPAUTH
static unsigned char authenticated=0;
#endif

void setup()
{
  char *x;
  unsigned long u;
 
#ifdef USE_SMTPAUTH
  extern struct commands smtpcommands[];
  struct commands	*p = smtpcommands;
#endif

  if (control_init() == -1) die_control();
  if (control_rldef(&greeting,"control/smtpgreeting",1,(char *) 0) != 1)
    die_control();
  liphostok = control_rldef(&liphost,"control/localiphost",1,(char *) 0);
  if (liphostok == -1) die_control();
  if (control_readint(&timeout,"control/timeoutsmtpd") == -1) die_control();
  if (timeout <= 0) timeout = 1;

  if (rcpthosts_init() == -1) die_control();
  if (spp_init() == -1) die_control();

  bmfok = control_readfile(&bmf,"control/badmailfrom",0);
  if (bmfok == -1) die_control();
  if (bmfok)
    if (!constmap_init(&mapbmf,bmf.s,bmf.len,0)) die_nomem();
 
  if (control_readint(&databytes,"control/databytes") == -1) die_control();
  x = env_get("DATABYTES");
  if (x) { scan_ulong(x,&u); databytes = u; }
  if (!(databytes + 1)) --databytes;
 
  if (control_readint(&spfbehavior,"control/spfbehavior") == -1)
    die_control();
  x = env_get("SPFBEHAVIOR");
  if (x) { scan_ulong(x,&u); spfbehavior = u; }

  if (control_readline(&spflocal,"control/spfrules") == -1) die_control();
  if (spflocal.len && !stralloc_0(&spflocal)) die_nomem();
  if (control_readline(&spfguess,"control/spfguess") == -1) die_control();
  if (spfguess.len && !stralloc_0(&spfguess)) die_nomem();
  if (control_rldef(&spfexp,"control/spfexp",0,SPF_DEFEXP) == -1)
    die_control();
  if (!stralloc_0(&spfexp)) die_nomem();

  remoteip = env_get("TCPREMOTEIP");
  if (!remoteip) remoteip = "unknown";
  local = env_get("TCPLOCALHOST");
  if (!local) local = env_get("TCPLOCALIP");
  if (!local) local = "unknown";
  remotehost = env_get("TCPREMOTEHOST");
  if (!remotehost) remotehost = "unknown";
  remoteinfo = env_get("TCPREMOTEINFO");
  relayclient = env_get("RELAYCLIENT");

#ifdef USE_SMTPAUTH
  if(relayclient)
	authenticated = 1;
  if(!(plesk_smtp_auth = env_get("SMTPAUTH")))
	for(;p->text; p++)
		if(!strcmp(p->text, "auth")) {
			p->fun = err_unimpl;
			break;
		}
#endif

#ifdef TLS
  if (env_get("SMTPS")) { smtps = 1; tls_init(); }
  else
#endif
  dohelo(remotehost);
}


stralloc addr = {0}; /* will be 0-terminated, if addrparse returns 1 */

int addrparse(arg)
char *arg;
{
  int i;
  char ch;
  char terminator;
  struct ip_address ip;
  int flagesc;
  int flagquoted;
 
  terminator = '>';
  i = str_chr(arg,'<');
  if (arg[i])
    arg += i + 1;
  else { /* partner should go read rfc 821 */
    terminator = ' ';
    arg += str_chr(arg,':');
    if (*arg == ':') ++arg;
    while (*arg == ' ') ++arg;
  }

  /* strip source route */
  if (*arg == '@') while (*arg) if (*arg++ == ':') break;

  if (!stralloc_copys(&addr,"")) die_nomem();
  flagesc = 0;
  flagquoted = 0;
  for (i = 0;ch = arg[i];++i) { /* copy arg to addr, stripping quotes */
    if (flagesc) {
      if (!stralloc_append(&addr,&ch)) die_nomem();
      flagesc = 0;
    }
    else {
      if (!flagquoted && (ch == terminator)) break;
      switch(ch) {
        case '\\': flagesc = 1; break;
        case '"': flagquoted = !flagquoted; break;
        default: if (!stralloc_append(&addr,&ch)) die_nomem();
      }
    }
  }
  /* could check for termination failure here, but why bother? */
  if (!stralloc_append(&addr,"")) die_nomem();

  if (liphostok) {
    i = byte_rchr(addr.s,addr.len,'@');
    if (i < addr.len) /* if not, partner should go read rfc 821 */
      if (addr.s[i + 1] == '[')
        if (!addr.s[i + 1 + ip_scanbracket(addr.s + i + 1,&ip)])
          if (ipme_is(&ip)) {
            addr.len = i + 1;
            if (!stralloc_cat(&addr,&liphost)) die_nomem();
            if (!stralloc_0(&addr)) die_nomem();
          }
  }

  if (addr.len > 900) return 0;
  return 1;
}

int bmfcheck()
{
  int j;
  if (!bmfok) return 0;
  if (constmap(&mapbmf,addr.s,addr.len - 1)) return 1;
  j = byte_rchr(addr.s,addr.len,'@');
  if (j < addr.len)
    if (constmap(&mapbmf,addr.s + j,addr.len - j - 1)) return 1;
  return 0;
}

int addrallowed()
{
  int r;
  r = rcpthosts(addr.s,str_len(addr.s));
  if (r == -1) die_control();
#ifdef TLS
  if (r == 0) if (tls_verify()) r = -2;
#endif
  return r;
}

int addrrelay()
{
  int j;
  j = addr.len;
  while(--j >= 0)
    if (addr.s[j] == '@') break;
  if (j < 0) j = addr.len;
  while(--j >= 0) {
    if (addr.s[j] == '@') return 1;
    if (addr.s[j] == '%') return 1;
    if (addr.s[j] == '!') return 1;
  }
  return 0;
}


int seenmail = 0;
int flagbarf; /* defined if seenmail */
int flagbarfspf;
int allowed;
stralloc spfbarfmsg = {0};
stralloc mailfrom = {0};
stralloc rcptto = {0};

int envelope_scanner()
{
  int child;
  int wstat;
  char *envelope_scannerarg[] = { "bin/qmail-envelope-scanner", mailfrom.s, addr.s, 0 };

  switch(child = vfork()) {
    case -1:
      return 1;
    case 0:
      execv(*envelope_scannerarg,envelope_scannerarg);
      _exit(111);
  }

  wait_pid(&wstat,child);
  if (wait_crashed(wstat)) {
    return 1;
  }

  switch(wait_exitcode(wstat)) {
    case 101:
      err_tempfail();
      return 0;
    case 100:
      err_permfail();
      return 0;
    default:
      return 1;
  }
}

void smtp_helo(arg) char *arg;
{
  if(!spp_helo(arg)) return;
  smtp_greet("250 "); out("\r\n");
  seenmail = 0; dohelo(arg);
}
void smtp_ehlo(arg) char *arg;
{
  if(!spp_helo(arg)) return;
  smtp_greet("250-");
#ifdef USE_SMTPAUTH
  if(plesk_smtp_auth && !authenticated) {

#ifdef USE_OLD_GREETING
#ifdef SMTPAUTH_NETSCAPE_WORKAROUND
  out("\r\n250-AUTH=LOGIN CRAM-MD5");
#else
  out("\r\n250-AUTH=LOGIN CRAM-MD5 PLAIN");
#endif
#endif

#ifdef USE_NEW_GREETING
#ifdef SMTPAUTH_NETSCAPE_WORKAROUND
  out("\r\n250-AUTH LOGIN CRAM-MD5");
#else
  out("\r\n250-AUTH LOGIN CRAM-MD5 PLAIN");
#endif
#endif
  }

#endif
#ifdef TLS
  if (!ssl) out("\r\n250-STARTTLS");
#endif
  out("\r\n250-PIPELINING\r\n250 8BITMIME\r\n");
  seenmail = 0; dohelo(arg);
}
void smtp_rset(char* arg)
{
  spp_rset();
  seenmail = 0;
  out("250 flushed\r\n");
}
void smtp_mail(arg) char *arg;
{
  int r;

  if (!addrparse(arg)) { err_syntax(); return; }
  if (!(spp_val = spp_mail())) return;
  if (spp_val == 1)
  flagbarf = bmfcheck();

  flagbarfspf = 0;
  if (spfbehavior && !relayclient)
     {
      switch(r = spfcheck()) {
      case SPF_OK: env_put2("SPFRESULT","pass"); break;
      case SPF_NONE: env_put2("SPFRESULT","none"); break;
      case SPF_UNKNOWN: env_put2("SPFRESULT","unknown"); break;
      case SPF_NEUTRAL: env_put2("SPFRESULT","neutral"); break;
      case SPF_SOFTFAIL: env_put2("SPFRESULT","softfail"); break;
      case SPF_FAIL: env_put2("SPFRESULT","fail"); break;
      case SPF_ERROR: env_put2("SPFRESULT","error"); break;
      }
      switch (r) {
      case SPF_NOMEM:
        die_nomem();
      case SPF_ERROR:
        if (spfbehavior < 2) break;
       out("451 SPF lookup failure (#4.3.0)\r\n");
       return;
      case SPF_NONE:
      case SPF_UNKNOWN:
        if (spfbehavior < 6) break;
      case SPF_NEUTRAL:
        if (spfbehavior < 5) break;
      case SPF_SOFTFAIL:
        if (spfbehavior < 4) break;
      case SPF_FAIL:
        if (spfbehavior < 3) break;
        if (!spfexplanation(&spfbarfmsg)) die_nomem();
        if (!stralloc_0(&spfbarfmsg)) die_nomem();
        flagbarfspf = 1;
     }
    }
   else
     env_unset("SPFRESULT");

  seenmail = 1;
  if (!stralloc_copys(&rcptto,"")) die_nomem();
  if (!stralloc_copys(&mailfrom,addr.s)) die_nomem();
  if (!stralloc_0(&mailfrom)) die_nomem();
  out("250 ok\r\n");
}
void err_spf() {
  int i,j;

  for(i = 0; i < spfbarfmsg.len; i = j + 1) {
     j = byte_chr(spfbarfmsg.s + i, spfbarfmsg.len - i, '\n') + i;
     if (j < spfbarfmsg.len) {
       out("550-");
       spfbarfmsg.s[j] = 0;
       out(spfbarfmsg.s);
       spfbarfmsg.s[j] = '\n';
       out("\r\n");
     } else {
       out("550 ");
       out(spfbarfmsg.s);
       out(" (#5.7.1)\r\n");
     }
  }
}
void smtp_rcpt(arg) char *arg; {
  if (!seenmail) { err_wantmail(); return; }
  if (!addrparse(arg)) { err_syntax(); return; }
  if (addrrelay()) { err_relay(); return; }
  if (flagbarf) { err_bmf(); return; }
  if (flagbarfspf) { err_spf(); return; }
  if (!relayclient) allowed = addrallowed();
  if (!(spp_val = spp_rcpt(allowed))) return;
  if (spp_val == 1) {
  if (relayclient) {
    --addr.len;
    if (!stralloc_cats(&addr,relayclient)) die_nomem();
    if (!stralloc_0(&addr)) die_nomem();
  }
  else {
    if (!allowed) { err_nogateway(); return; }
    if (!envelope_scanner()) return;
   }
  }
  spp_rcpt_accepted();
    
  if (!stralloc_cats(&rcptto,"T")) die_nomem();
  if (!stralloc_cats(&rcptto,addr.s)) die_nomem();
  if (!stralloc_0(&rcptto)) die_nomem();
  out("250 ok\r\n");
}
int saferead(fd,buf,len) int fd; char *buf; int len;
{
  int r;
  flush();
#ifdef TLS
  if (ssl && fd == ssl_rfd)
    r = ssl_timeoutread(timeout, ssl_rfd, ssl_wfd, ssl, buf, len);
  else
#endif
  r = timeoutread(timeout,fd,buf,len);
  if (r == -1) if (errno == error_timeout) die_alarm();
  if (r <= 0) die_read();
  return r;
}

char ssinbuf[1024];
substdio ssin = SUBSTDIO_FDBUF(saferead,0,ssinbuf,sizeof ssinbuf);

struct qmail qqt;
unsigned int bytestooverflow = 0;

void put(ch)
char *ch;
{
  if (bytestooverflow)
    if (!--bytestooverflow)
      qmail_fail(&qqt);
  qmail_put(&qqt,ch,1);
}

void blast(hops)
int *hops;
{
  char ch;
  int state;
  int flaginheader;
  int pos; /* number of bytes since most recent \n, if fih */
  int flagmaybex; /* 1 if this line might match RECEIVED, if fih */
  int flagmaybey; /* 1 if this line might match \r\n, if fih */
  int flagmaybez; /* 1 if this line might match DELIVERED, if fih */
 
  state = 1;
  *hops = 0;
  flaginheader = 1;
  pos = 0; flagmaybex = flagmaybey = flagmaybez = 1;
  for (;;) {
    substdio_get(&ssin,&ch,1);
    if (flaginheader) {
      if (pos < 9) {
        if (ch != "delivered"[pos]) if (ch != "DELIVERED"[pos]) flagmaybez = 0;
        if (flagmaybez) if (pos == 8) ++*hops;
        if (pos < 8)
          if (ch != "received"[pos]) if (ch != "RECEIVED"[pos]) flagmaybex = 0;
        if (flagmaybex) if (pos == 7) ++*hops;
        if (pos < 2) if (ch != "\r\n"[pos]) flagmaybey = 0;
        if (flagmaybey) if (pos == 1) flaginheader = 0;
        ++pos;
      }
      if (ch == '\n') { pos = 0; flagmaybex = flagmaybey = flagmaybez = 1; }
    }
    switch(state) {
      case 0:
        if (ch == '\n') straynewline();
        if (ch == '\r') { state = 4; continue; }
        break;
      case 1: /* \r\n */
        if (ch == '\n') straynewline();
        if (ch == '.') { state = 2; continue; }
        if (ch == '\r') { state = 4; continue; }
        state = 0;
        break;
      case 2: /* \r\n + . */
        if (ch == '\n') straynewline();
        if (ch == '\r') { state = 3; continue; }
        state = 0;
        break;
      case 3: /* \r\n + .\r */
        if (ch == '\n') return;
        put(".");
        put("\r");
        if (ch == '\r') { state = 4; continue; }
        state = 0;
        break;
      case 4: /* + \r */
        if (ch == '\n') { state = 1; break; }
        if (ch != '\r') { put("\r"); state = 0; }
    }
    put(&ch);
  }
}

void spfreceived()
{
  stralloc sa = {0};
  stralloc rcvd_spf = {0};

  if (!spfbehavior || relayclient) return;

  if (!stralloc_copys(&rcvd_spf, "Received-SPF: ")) die_nomem();
  if (!spfinfo(&sa)) die_nomem();
  if (!stralloc_cat(&rcvd_spf, &sa)) die_nomem();
  if (!stralloc_append(&rcvd_spf, "\n")) die_nomem();
  if (bytestooverflow) {
    bytestooverflow -= rcvd_spf.len;
    if (bytestooverflow <= 0) qmail_fail(&qqt);
  }
  qmail_put(&qqt,rcvd_spf.s,rcvd_spf.len);
}


char accept_buf[FMT_ULONG];
void acceptmessage(qp) unsigned long qp;
{
  datetime_sec when;
  when = now();
  out("250 ok ");
  accept_buf[fmt_ulong(accept_buf,(unsigned long) when)] = 0;
  out(accept_buf);
  out(" qp ");
  accept_buf[fmt_ulong(accept_buf,qp)] = 0;
  out(accept_buf);
  out("\r\n");
}

void smtp_data(char* arg) {
  int hops;
  unsigned long qp;
  char *qqx;
 
  if (!seenmail) { err_wantmail(); return; }
  if (!rcptto.len) { err_wantrcpt(); return; }
  if (!spp_data()) return;
  seenmail = 0;
  if (databytes) bytestooverflow = databytes + 1;
  if (qmail_open(&qqt) == -1) { err_qqt(); return; }
  qp = qmail_qp(&qqt);
  out("354 go ahead\r\n");
 
  received(&qqt,protocol,local,remoteip,remotehost,remoteinfo,fakehelo);
  spfreceived();
  qmail_put(&qqt,sppheaders.s,sppheaders.len); /* set in qmail-spp.c */
  spp_rset();
  blast(&hops);
  hops = (hops >= MAXHOPS);
  if (hops) qmail_fail(&qqt);
  qmail_from(&qqt,mailfrom.s);
  qmail_put(&qqt,rcptto.s,rcptto.len);
 
  qqx = qmail_close(&qqt);
  if (!*qqx) { acceptmessage(qp); return; }
  if (hops) { out("554 too many hops, this message is looping (#5.4.6)\r\n"); return; }
  if (databytes) if (!bytestooverflow) { out("552 sorry, that message size exceeds my databytes limit (#5.3.4)\r\n"); return; }
  if (*qqx == 'D') out("554 "); else out("451 ");
  out(qqx + 1);
  out("\r\n");
}

#ifdef TLS
stralloc proto = {0};
int ssl_verified = 0;
const char *ssl_verify_err = 0;

void smtp_tls(char *arg)
{
  if (ssl) err_unimpl(0);
  else if (*arg) out("501 Syntax error (no parameters allowed) (#5.5.4)\r\n");
  else tls_init();
}

RSA *tmp_rsa_cb(SSL *ssl, int export, int keylen)
{
  if (!export) keylen = 512;
  if (keylen == 512) {
    FILE *in = fopen("control/rsa512.pem", "r");
    if (in) {
      RSA *rsa = PEM_read_RSAPrivateKey(in, NULL, NULL, NULL);
      fclose(in);
      if (rsa) return rsa;
    }
  }
  return RSA_generate_key(keylen, RSA_F4, NULL, NULL);
}

DH *tmp_dh_cb(SSL *ssl, int export, int keylen)
{
  if (!export) keylen = 1024;
  if (keylen == 512) {
    FILE *in = fopen("control/dh512.pem", "r");
    if (in) {
      DH *dh = PEM_read_DHparams(in, NULL, NULL, NULL);
      fclose(in);
      if (dh) return dh;
    }
  }
  if (keylen == 1024) {
    FILE *in = fopen("control/dh1024.pem", "r");
    if (in) {
      DH *dh = PEM_read_DHparams(in, NULL, NULL, NULL);
      fclose(in);
      if (dh) return dh;
    }
  }
  return DH_generate_parameters(keylen, DH_GENERATOR_2, NULL, NULL);
} 

/* don't want to fail handshake if cert isn't verifiable */
int verify_cb(int preverify_ok, X509_STORE_CTX *ctx) { return 1; }

void tls_nogateway()
{
  /* there may be cases when relayclient is set */
  if (!ssl || relayclient) return;
  out("; no valid cert for gatewaying");
  if (ssl_verify_err) { out(": "); out(ssl_verify_err); }
}
void tls_out(const char *s1, const char *s2)
{
  out("454 TLS "); out(s1);
  if (s2) { out(": "); out(s2); }
  out(" (#4.3.0)\r\n"); flush();
}
void tls_err(const char *s) { tls_out(s, ssl_error()); if (smtps) die_read(); }

# define CLIENTCA "control/clientca.pem"
# define CLIENTCRL "control/clientcrl.pem"
# define SERVERCERT "control/servercert.pem"

int tls_verify()
{
  stralloc clients = {0};
  struct constmap mapclients;

  if (!ssl || relayclient || ssl_verified) return 0;
  ssl_verified = 1; /* don't do this twice */

  /* request client cert to see if it can be verified by one of our CAs
   * and the associated email address matches an entry in tlsclients */
  switch (control_readfile(&clients, "control/tlsclients", 0))
  {
    case 1:
      if (constmap_init(&mapclients, clients.s, clients.len, 0)) {
        /* if CLIENTCA contains all the standard root certificates, a
         * 0.9.6b client might fail with SSL_R_EXCESSIVE_MESSAGE_SIZE;
         * it is probably due to 0.9.6b supporting only 8k key exchange
         * data while the 0.9.6c release increases that limit to 100k */
        STACK_OF(X509_NAME) *sk = SSL_load_client_CA_file(CLIENTCA);
        if (sk) {
          SSL_set_client_CA_list(ssl, sk);
          SSL_set_verify(ssl, SSL_VERIFY_PEER | SSL_VERIFY_CLIENT_ONCE, NULL);
          break;
        }
        constmap_free(&mapclients);
      }
    case 0: alloc_free(clients.s); return 0;
    case -1: die_control();
  }

  if (ssl_timeoutrehandshake(timeout, ssl_rfd, ssl_wfd, ssl) <= 0) {
    const char *err = ssl_strerror();
    tls_out("rehandshake failed", err); die_read();
  }

  do { /* one iteration */
    X509 *peercert;
    X509_NAME *subj;
    stralloc email = {0};

    int n = SSL_get_verify_result(ssl);
    if (n != X509_V_OK)
      { ssl_verify_err = X509_verify_cert_error_string(n); break; }
    peercert = SSL_get_peer_certificate(ssl);
    if (!peercert) break;

    subj = X509_get_subject_name(peercert);
    n = X509_NAME_get_index_by_NID(subj, NID_pkcs9_emailAddress, -1);
    if (n >= 0)
    {
      const ASN1_STRING *s = X509_NAME_get_entry(subj, n)->value;

      // CVB 2007-07-28: Refactoring
      if (s)
      {
        email.len = s->length;
        email.s = (char *)s->data;
      }
    }

    if (email.len <= 0)
      ssl_verify_err = "contains no email address";
    else if (!constmap(&mapclients, email.s, email.len))
      ssl_verify_err = "email address not in my list of tlsclients";
    else {
      /* add the cert email to the proto if it helped allow relaying */
      --proto.len;
      if (!stralloc_cats(&proto, "\n  (cert ") /* continuation line */
        || !stralloc_catb(&proto, email.s, email.len)
        || !stralloc_cats(&proto, ")")
        || !stralloc_0(&proto)) die_nomem();
      relayclient = "";
      protocol = proto.s;
    }

    X509_free(peercert);
  } while (0);
  constmap_free(&mapclients); alloc_free(clients.s);

  /* we are not going to need this anymore: free the memory */
  SSL_set_client_CA_list(ssl, NULL);
  SSL_set_verify(ssl, SSL_VERIFY_NONE, NULL);

  return relayclient ? 1 : 0;
}

void tls_init()
{
  SSL *myssl;
  SSL_CTX *ctx;
  const char *ciphers;
  stralloc saciphers = {0};
  X509_STORE *store;
  X509_LOOKUP *lookup;

  SSL_library_init();

  /* a new SSL context with the bare minimum of options */
  ctx = SSL_CTX_new(SSLv23_server_method());
  if (!ctx) { tls_err("unable to initialize ctx"); return; }

  if (!SSL_CTX_use_certificate_chain_file(ctx, SERVERCERT))
    { SSL_CTX_free(ctx); tls_err("missing certificate"); return; }
  SSL_CTX_load_verify_locations(ctx, CLIENTCA, NULL);

#if OPENSSL_VERSION_NUMBER >= 0x00907000L
  /* crl checking */
  store = SSL_CTX_get_cert_store(ctx);
  if ((lookup = X509_STORE_add_lookup(store, X509_LOOKUP_file())) &&
      (X509_load_crl_file(lookup, CLIENTCRL, X509_FILETYPE_PEM) == 1))
    X509_STORE_set_flags(store, X509_V_FLAG_CRL_CHECK |
                                X509_V_FLAG_CRL_CHECK_ALL);
#endif

  /* set the callback here; SSL_set_verify didn't work before 0.9.6c */
  SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, verify_cb);

  /* a new SSL object, with the rest added to it directly to avoid copying */
  myssl = SSL_new(ctx);
  SSL_CTX_free(ctx);
  if (!myssl) { tls_err("unable to initialize ssl"); return; }

  /* this will also check whether public and private keys match */
  if (!SSL_use_RSAPrivateKey_file(myssl, SERVERCERT, SSL_FILETYPE_PEM))
    { SSL_free(myssl); tls_err("no valid RSA private key"); return; }

  ciphers = env_get("TLSCIPHERS");
  if (!ciphers) {
    if (control_readfile(&saciphers, "control/tlsserverciphers") == -1)
      { SSL_free(myssl); die_control(); }
    if (saciphers.len) { /* convert all '\0's except the last one to ':' */
      int i;
      for (i = 0; i < saciphers.len - 1; ++i)
        if (!saciphers.s[i]) saciphers.s[i] = ':';
      ciphers = saciphers.s;
    }
  }
  if (!ciphers || !*ciphers) ciphers = "DEFAULT";
  SSL_set_cipher_list(myssl, ciphers);
  alloc_free(saciphers.s);

  SSL_set_tmp_rsa_callback(myssl, tmp_rsa_cb);
  SSL_set_tmp_dh_callback(myssl, tmp_dh_cb);
  SSL_set_rfd(myssl, ssl_rfd = substdio_fileno(&ssin));
  SSL_set_wfd(myssl, ssl_wfd = substdio_fileno(&ssout));

  if (!smtps) { out("220 ready for tls\r\n"); flush(); }

  if (ssl_timeoutaccept(timeout, ssl_rfd, ssl_wfd, myssl) <= 0) {
    /* neither cleartext nor any other response here is part of a standard */
    const char *err = ssl_strerror();
    ssl_free(myssl); tls_out("connection failed", err); die_read();
  }
  ssl = myssl;

  /* populate the protocol string, used in Received */
  if (!stralloc_copys(&proto, "(")
    || !stralloc_cats(&proto, SSL_get_cipher(ssl))
    || !stralloc_cats(&proto, " encrypted) SMTP")) die_nomem();
  if (!stralloc_0(&proto)) die_nomem();
  protocol = proto.s;

  /* have to discard the pre-STARTTLS HELO/EHLO argument, if any */
  dohelo(remotehost);
}

# undef SERVERCERT
# undef CLIENTCA

#endif

// CVB 2007-07-28: Refactoring
#ifdef USE_SMTPAUTH
/* static unsigned char authenticated=0; */
static unsigned char *base64_alphabet =
(unsigned char *)"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

static int unbase64(ch) int ch;
{
  int i;
  if (ch == '=') return 0;
  for (i = 0; i < 64; i++)
    if (ch == base64_alphabet[i])
      return i;
  return 0;
}

// CVB 2007-07-28: Refactoring
static int base64_dec_buffer(str, dst, len) const char *str; void *dst; int len;
{
  int i, j, l;
  unsigned char input[4], output[3], *result = (unsigned char *)dst;

  if (str == 0)
  {
    return 0;
  }

  l = str_len(str);

  if (dst == 0 || l > len)
  {
     return (l / 4) * 3;
  }

  memset(dst,0,len);

  for (i=j=0; i<l; i+=4)
  {
    input[0] = unbase64(str[i]);
    input[1] = unbase64(str[i+1]);
    input[2] = unbase64(str[i+2]);
    input[3] = unbase64(str[i+3]);
    output[0] = (input[0] << 2) | (input[1] >> 4);
    output[1] = (input[1] << 4) | (input[2] >> 2);
    output[2] = (input[2] << 6) | (input[3]);
    result[j] = output[0];
    if (str[i+1] == '=') return j+1;
    result[j+1]=output[1];
    if (str[i+2] == '=') return j+2;
    result[j+2]=output[2];
    j += 3;
  }

  return j;
}

static char **smtpauth_argv;
static char *auth_argv[4];
static stralloc smtpauth = {0};
static char smtpauthlogin[65];
static char smtpauthpass[65];
static char smtpauthtimestamp[65];

char unique[FMT_ULONG + FMT_ULONG + 3];

void base64encode(stralloc *input, stralloc *output)
{
int	a=0,b=0,c=0;
int	i, j;
int	d, e, f, g;

	if (input->len == 0)	return;

	for (j=i=0; i< input->len ; i += 3)
	{
		a=input->s[i];
		b= i+1 < input->len ? input->s[i+1]:0;
		c= i+2 < input->len ? input->s[i+2]:0;

		d=base64_alphabet[ a >> 2 ];
		e=base64_alphabet[ ((a & 3 ) << 4) | (b >> 4)];
		f=base64_alphabet[ ((b & 15) << 2) | (c >> 6)];
		g=base64_alphabet[ c & 63 ];
		if (i + 1 >= input->len) f='=';
		if (i + 2 >= input->len) g='=';
		stralloc_append(output,&d);
		stralloc_append(output,&e);
		stralloc_append(output,&f);
		stralloc_append(output,&g);
	}
}

static int smtpauth_getl(void) {
  int i;
  if (!stralloc_copys(&smtpauth, "")) return -1;
  for (;;) {
    if (!stralloc_readyplus(&smtpauth,1)) return -1;
    i = substdio_get(&ssin, smtpauth.s + smtpauth.len, 1);
    if (i != 1) return i;
    if (smtpauth.s[smtpauth.len] == '\n') break;
    ++smtpauth.len;
  }
  if (smtpauth.len > 0) if (smtpauth.s[smtpauth.len-1] == '\r') --smtpauth.len;
  smtpauth.s[smtpauth.len] = 0;
  return smtpauth.len;
}

static void smtpauth_authenticate(void)
{
  int st, pid, fds[2];
  
    if (pipe(fds)) {
      out("535 pipe failure\r\n");
      flush();
      _exit(0);
    }
    /* spawn external program

    external program should return '0' if it was successful,

    submit: /bin/checkpassword /bin/true
  
    */
    switch ((pid=fork())) {
      case -1: die_nomem();
      case 0: close(fds[1]);
        fd_copy(3,fds[0]);
        flush();
        execvp(auth_argv[1], auth_argv+1);
        die_nomem();
    };
    close(fds[0]);
    write(fds[1], smtpauthlogin, str_len(smtpauthlogin)+1);
    write(fds[1], smtpauthpass, str_len(smtpauthpass)+1);
    if (str_len(smtpauthtimestamp))
    {
          write(fds[1], smtpauthtimestamp, str_len(smtpauthtimestamp)+1);
    }
    close(fds[1]);
    wait_pid(&st, pid);
    if (wait_exitcode(st) == 0) {
      out("235 go ahead\r\n");
      flush();
      relayclient="";
      authenticated=1;
      return;
    }
    sleep(2);
    out("535 auth failure\r\n"); 
    flush(); 
    return;
}

void smtp_auth(arg) char *arg; {
  int ret,i,toolong,start;
  char *helper;
  /* netscape 4.5 sends AUTH LOGIN <base64encodedusername>
     microsoft outlook express sends AUTH LOGIN

     idea is simple

     use an external program to test authority
     if success, set 'RELAYCLIENT'
     otherwise, let them know nicely (hangup)

     note, i really don't like djb's coding style even though i'm using it here.
     i think using spaces for tabs is bad.
                                                -mrs.brisby@nimh.org
  */
  /* Here i've added support for other auth types.
  
  						-brush@elysium.pl	*/
  if (!authenticated)
  {
    if ((ret=strncasecmp(arg,"login",5))==0)
    {
      while (arg && *arg && *arg != ' ') arg++;

      /* pass over the space */
      while (arg && *arg && *arg == ' ') arg++;

      if (arg && *arg) {
        /* here's the base64 encoded login */
        base64_dec_buffer(arg, smtpauthlogin, sizeof(smtpauthlogin));
      } else {
        out("334 VXNlcm5hbWU6\r\n"); /* b64 <- 'Username:' */
        flush();
        if (smtpauth_getl() > 0)
          base64_dec_buffer(smtpauth.s, smtpauthlogin, sizeof(smtpauthlogin));
        else
          die_read();
      }
      out("334 UGFzc3dvcmQ6\r\n"); /* b64 <- 'Password:' */
      flush();
      if (smtpauth_getl() > 0)
        base64_dec_buffer(smtpauth.s, smtpauthpass, sizeof(smtpauthpass));
      else
        die_read();
      smtpauthtimestamp[0]=0;
      auth_argv[1]=smtpauth_argv[1];	/* change checkpass prg */
      auth_argv[2]=smtpauth_argv[2];	/* change checkpass prg */
      auth_argv[3]=NULL;			/* change checkpass prg */
      smtpauth_authenticate();  
      return;
    }
    else if ((ret=strncasecmp(arg,"plain",5))==0)  
    {
      // check for not obligatory argument - initial-response, rfc2554
      // AUTH mechanism [initial-response]
      int init_response_is_used;
	  int len;

	  // CHECK FOR initial-response
	  i = 5;
	  len = strlen(&arg[i]); // first 5 chars is plain check chars after ...
	  while ((0 < len) && (' ' == arg[i])) {i++; len--;}
      // (len >= 4) means that auth_id\0user_id\0passwd should be specified, auth_id can be empty
	  init_response_is_used = (len >= 4);
	  start = i;

	  if (init_response_is_used) {
		// DECODE base64 string
        char buff[sizeof(smtpauthlogin)+sizeof(smtpauthlogin)+sizeof(smtpauthpass)+2];
        len = base64_dec_buffer(&arg[start], buff, sizeof(buff));
		stralloc_copyb(&smtpauth, buff, len);
      } else {
	    // ASK FOR auth line auth_id\0user_id\0passwd, auth_id can be empty 
        out("334 ok. go on.\r\n");
        flush();
      }

      if (init_response_is_used || (smtpauth_getl() > 0))
      {
		// SKIP auth_id and first null delimiter
		i=0;
        for (;0!=smtpauth.s[i] && i<sizeof(smtpauthlogin) && i<smtpauth.len;i++);
		if (0==smtpauth.s[i] && i<smtpauth.len)
			i++;

		// READ user_id
        toolong=1;
		start = i;
        for (;i<(sizeof(smtpauthlogin)+start) && i<smtpauth.len;i++)
        {
          smtpauthlogin[i-start]=smtpauth.s[i];
          if(smtpauth.s[i]==0)
          {
            toolong=0;
            break;
          }
        }
        start=i;
        if (toolong==1)
        {
          out("535 malformed input\r\n"); 
          flush(); 
          return;
        }

		// READ passwd
        toolong=1;
        i++;
        start=i;
        for (; i<(sizeof(smtpauthpass)+start) && i<smtpauth.len;i++)
        {
          smtpauthpass[i-start]=smtpauth.s[i];
          if(smtpauth.s[i]==0)
          {
            toolong=0;
            break;
          }
        }
		if (i<(sizeof(smtpauthpass)+start) && (i == smtpauth.len)) {
		  // auth string may haven't NULL char at the end, it is normal
          smtpauthpass[i-start]=0;
          toolong=0;
		}

        start=i;
        if (toolong==1)
        {
          out("535 malformed input\r\n"); 
          flush(); 
          return;
        }
        smtpauthtimestamp[0]=0;
        auth_argv[1]=smtpauth_argv[1];		/* change checkpass prg */
        auth_argv[2]=smtpauth_argv[2];		/* change checkpass prg */
        auth_argv[3]=NULL;			/* change checkpass prg */
        smtpauth_authenticate();
        return;
      }
      else
      {
        die_read();
      }
    }
    else if ((ret=strncasecmp(arg,"cram-md5",8))==0)  
    {
      int r;
      static stralloc me = {0};
      static stralloc greet = {0};
      static stralloc greetenc = {0};
      char *s;
      r = control_readline(&me,"control/me");
      if (r != 1) 
      { out("535 internal server error\r\n"); flush(); _exit(0);
      }
      for (r=0;r<me.len;r++)
      {
        if (me.s[r]=='\n' || me.s[r]=='\r')
        {
          me.s[r]=0;
          break;
        }
      }
      me.s[me.len]=0; // obscure fix but it works
      s = unique;
      s += fmt_uint(s,getpid());
      *s++ = '.';
      s += fmt_ulong(s,(unsigned long) now());
      *s++ = '@';
      *s++ = 0;
      if (greet.len)
      {
	for (i=0;i<greet.len;i++)
	{
	  greet.s[i]=0;
	}
        greet.len=0;
      }
      stralloc_append(&greet,"<");
      stralloc_cats(&greet,unique);
      stralloc_cats(&greet,me.s);
      stralloc_cats(&greet,">");
      greet.s[greet.len]=0;	// obscure fix but it works
      stralloc_readyplus(&greet,3);
      if (greetenc.len)
      {
	for (i=0;i<greetenc.len;i++)
	{
	  greetenc.s[i]=0;
	}
        greetenc.len=0;
      }
      flush();
      base64encode(&greet,&greetenc);
      greetenc.s[greetenc.len]=0; // obscure fix but it works
      out("334 ");
      out(greetenc.s);
      out("\r\n");
      flush();
      if (smtpauth_getl() > 0)
      {
        s=calloc((size_t) strlen(smtpauth.s),(size_t)1);
        base64_dec_buffer(smtpauth.s, s, strlen(smtpauth.s));
      }
      helper=strtok(s," ");
      if(helper!=NULL)
      {
        strncpy (smtpauthlogin,helper,64);
      }
      else
      {
        out("535 malformed input\r\n"); 
        return;
      }
      helper=strtok(NULL," ");
      if(helper!=NULL)
      {
        strncpy (smtpauthtimestamp,helper,64);
      }
      else
      {
        out("535 malformed input\r\n"); 
        return;
      }
      strncpy (smtpauthpass,greet.s,64);
      auth_argv[1]=smtpauth_argv[3];	/* change checkpass prg */
      auth_argv[2]=smtpauth_argv[4];	/* change checkpass prg */
      auth_argv[3]=NULL;		/* change checkpass prg */
      smtpauth_authenticate();
      return;
    }
    else
    {
      out("504 auth type not supported\r\n"); 
      flush();
      return;
    }
  }
  else
  {
      out("503 you are already authenticated\r\n"); 
      flush(); 
      return;
  }
}
#endif

struct commands smtpcommands[] = {
  { "rcpt", smtp_rcpt, 0 }
, { "mail", smtp_mail, 0 }
, { "data", smtp_data, flush }
#ifdef USE_SMTPAUTH
, { "auth", smtp_auth, flush }
#endif
, { "quit", smtp_quit, flush }
, { "helo", smtp_helo, flush }
, { "ehlo", smtp_ehlo, flush }
, { "rset", smtp_rset, 0 }
, { "help", smtp_help, flush }
#ifdef TLS
, { "starttls", smtp_tls, flush }
#endif
, { "noop", err_noop, flush }
, { "vrfy", err_vrfy, flush }
, { 0, err_unimpl, flush }
} ;

int main(argc,argv)
int argc;
char **argv;
{
#ifdef USE_SMTPAUTH
  smtpauth_argv = argv;
#endif
  sig_pipeignore();
  if (chdir(auto_qmail) == -1) die_control();
  setup();
  if (ipme_init() != 1) die_ipme();
  if (spp_connect()) {
  smtp_greet("220 ");
  out(" ESMTP\r\n");
  }
  if (commands(&ssin,&smtpcommands) == 0) die_read();
  die_nomem();
}
